"""
Various useful functions


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import numpy as np
import scipy.linalg as LA
import sys


__all__ = ['sumsq', 'eval_least_squares_with_regularisation', 'model_value', 'random_orthog_directions_within_bounds',
           'random_directions_within_bounds', 'apply_scaling', 'remove_scaling', 'pbox', 'pball', 'dykstra', 'qr_rank']

module_logger = logging.getLogger(__name__) 


def sumsq(x):
    # There are several ways to calculate sum of squares of a vector:
    #   np.dot(x,x)
    #   np.sum(x**2)
    #   np.sum(np.square(x))
    #   etc.
    # Using the timeit routine, it seems like dot(x,x) is ~3-4x faster than other methods
    return np.dot(x, x)


def eval_least_squares_with_regularisation(objfun, x, h=None, argsf=(), argsh=(), verbose=True, eval_num=0, pt_num=0, full_x_thresh=6, check_for_overflow=True):
    # Evaluate least squares function
    fvec = objfun(x, *argsf)

    if check_for_overflow:
        try:
            if np.max(np.abs(fvec)) >= np.sqrt(sys.float_info.max):
                f = sys.float_info.max
            else:
                f = sumsq(fvec)  # objective = sum(ri^2) [no 1/2 factor at front]
        except OverflowError:
            f = sys.float_info.max
    else:
        f = sumsq(fvec)

    # objective = least-squares + regularisation
    obj = f
    if h is not None:
        # Evaluate regularisation term
        hvalue = h(x, *argsh)
        obj = f + hvalue

    if verbose:
        if len(x) < full_x_thresh:
            module_logger.info("Function eval %i at point %i has obj = %.15g at x = " % (eval_num, pt_num, obj) + str(x))
        else:
            module_logger.info("Function eval %i at point %i has obj = %.15g at x = [...]" % (eval_num, pt_num, obj))

    return fvec, obj


def model_value(g, H, s, xopt=(), h=None,argsh=(), scaling_changes=None):
    # Calculate model value (s^T * g + 0.5* s^T * H * s) + h(xopt + s) = s^T * (gopt + 0.5 * H*s) + h(xopt + s)
    assert g.shape == s.shape, "g and s have incompatible sizes"
    Hs = H.dot(s)
    rtn = np.dot(s, g + 0.5*Hs)
    if h is not None:
        hvalue = h(remove_scaling(xopt+s, scaling_changes), *argsh)
        rtn += hvalue
    return rtn


def get_scale(dirn, delta, lower, upper):
    scale = delta
    for j in range(len(dirn)):
        if dirn[j] < 0.0:
            scale = min(scale, lower[j] / dirn[j])
        elif dirn[j] > 0.0:
            scale = min(scale, upper[j] / dirn[j])
    return scale


def random_orthog_directions_within_bounds(num_pts, delta, lower, upper, with_neg_dirns=True):
    # Generate num_pts random directions d1, d2, ...
    # so that lower <= d1 <= upper and ||d1|| ~ delta [perhaps not equal if constraint active]
    # Try to encourage a 'star shape' of orthogonal directions first
    # then completely random after that.
    n = len(lower)
    assert lower.shape == (n,), "lower must be a vector"
    assert upper.shape == (n,), "lower and upper have incompatible sizes"
    assert np.min(upper) >= -1e-15, "upper must be non-negative"
    assert np.max(lower) <= 1e-15, "lower must be non-positive"
    assert np.min(upper - lower) > 0.0, "upper must be > lower"
    assert delta > 0, "delta must be strictly positive"
    assert num_pts > 0, "num_pts must be strictly positive"
    if with_neg_dirns:
        results = np.zeros((n, max(2*n, num_pts)))  # save space for results
    else:
        results = np.zeros((n, max(n, num_pts)))  # save space for results
    # Find the active set
    idx_l = (lower == 0)
    idx_u = (upper == 0)
    active = np.logical_or(idx_l, idx_u)
    inactive = np.logical_not(active)
    nactive = np.sum(active)
    ninactive = n - nactive
    # Get a random orthonormal basis for the inactive variables
    if ninactive > 0:
        A = np.random.normal(size=(ninactive, ninactive))
        Qred = np.linalg.qr(A)[0]  # orthonormal columns (reduced)
        Q = np.zeros((n, ninactive))  # full set of orthonormal vectors
        Q[inactive, :] = Qred  # zero change for active variables
        # 1. Orthogonal directions
        for i in range(ninactive):
            scale = get_scale(Q[:,i], delta, lower, upper)
            results[:, i] = scale * Q[:, i]
    # 2. Directions for active constraints
    idx_active = np.where(active)[0]  # indices of active constraints
    for i in range(nactive):
        idx = idx_active[i]
        results[idx, ninactive+i] = 1.0 if idx_l[idx] else -1.0
        results[:, ninactive+i] = get_scale(results[:, ninactive+i], delta, lower, upper) * results[:, ninactive+i]
    # 3. Negative orthogonal directions
    if with_neg_dirns:
        for i in range(ninactive):
            scale = get_scale(-Q[:, i], delta, lower, upper)
            results[:, n+i] = -scale * Q[:, i]
        # 4. Extra directions for active constraints
        for i in range(nactive):
            idx = idx_active[i]
            sign = 1.0 if idx_l[idx] else -1.0  # desired sign of direction shift
            if upper[idx] - lower[idx] > delta:
                results[idx, n+ninactive+i] = 2.0 * sign * delta
            else:
                results[idx, n + ninactive + i] = 0.5 * sign * (upper[idx] - lower[idx])
                # To get correct scaling, don't use delta any more (too big), use the scaling as given by upper-lower
            results[:, n+ninactive+i] = get_scale(results[:, n+ninactive+i], 1.0, lower, upper)*results[:, n+ninactive+i]
    # 5. Pad out the rest with random extra directions
    for i in range(num_pts - (2*n if with_neg_dirns else n)):
        dirn = np.random.normal(size=(n,))
        for j in range(nactive):
            idx = idx_active[j]
            sign = 1.0 if idx_l[idx] else -1.0  # desired sign of direction shift
            if dirn[idx]*sign < 0.0:
                dirn[idx] *= -1.0
        dirn = dirn / np.linalg.norm(dirn)
        scale = get_scale(dirn, delta, lower, upper)
        results[:, (2*n if with_neg_dirns else n)+i] = dirn * scale
    # Finally, make sure everything is within bounds
    for i in range(num_pts):
        results[:, i] = np.maximum(np.minimum(results[:, i], upper), lower)
    return results[:, :num_pts].T


def random_directions_within_bounds(num_pts, delta, lower, upper):
    # Generate num_pts random directions d1, d2, ...
    # so that lower <= d1 <= upper and ||d1|| ~ delta [perhaps not equal if constraint active]
    # Directions should be completely random (as much as possible while staying within bounds)
    n = len(lower)
    assert lower.shape == (n,), "lower must be a vector"
    assert upper.shape == (n,), "lower and upper have incompatible sizes"
    assert np.min(upper) >= -1e-15, "upper must be non-negative"
    assert np.max(lower) <= 1e-15, "lower must be non-positive"
    assert np.min(upper - lower) > 0.0, "upper must be > lower"
    assert delta > 0, "delta must be strictly positive"
    assert num_pts > 0, "num_pts must be strictly positive"
    results = np.zeros((n, num_pts))  # save space for results
    # Find the active set
    idx_l = (lower == 0)
    idx_u = (upper == 0)
    active = np.logical_or(idx_l, idx_u)
    # inactive = np.logical_not(active)
    nactive = np.sum(active)
    # ninactive = n - nactive
    idx_active = np.where(active)[0]  # indices of active constraints
    for i in range(num_pts):
        dirn = np.random.normal(size=(n,))
        for j in range(nactive):
            idx = idx_active[j]
            sign = 1.0 if idx_l[idx] else -1.0  # desired sign of direction shift
            if dirn[idx]*sign < 0.0:
                dirn[idx] *= -1.0
        dirn = dirn / np.linalg.norm(dirn)
        scale = get_scale(dirn, delta, lower, upper)
        results[:, i] = dirn * scale
    # Finally, scale by delta and make sure everything is within bounds
    for i in range(num_pts):
        results[:, i] = np.maximum(np.minimum(results[:, i], upper), lower)
    return results.T


def apply_scaling(x_raw, scaling_changes):
    if scaling_changes is None:
        return x_raw
    shift, scale = scaling_changes
    return (x_raw - shift) / scale


def remove_scaling(x_scaled, scaling_changes):
    if scaling_changes is None:
        return x_scaled
    shift, scale = scaling_changes
    return shift + x_scaled * scale


def dykstra(P,x0,max_iter=100,tol=1e-10):
    x = x0.copy()
    p = len(P)
    y = np.zeros((p,x0.shape[0]))

    n = 0
    cI = float('inf')
    while n < max_iter and cI >= tol:
        cI = 0
        for i in range(0,p):
            # Update iterate
            prev_x = x.copy()
            x = P[i](prev_x - y[i,:])

            # Update increment
            prev_y = y[i,:].copy()
            y[i,:] = x - (prev_x - prev_y)

            # Stop condition
            cI += np.linalg.norm(prev_y - y[i,:])**2

        n += 1

    return x


def pball(x,c,r):
    return c + (r/np.max([np.linalg.norm(x-c),r]))*(x-c)


def pbox(x,l,u):
    return np.minimum(np.maximum(x,l), u)

'''
Calculates rank of square matrix with QR.
We use the fact that the rank of a square matrix A
can be given by the number of nonzero diagonal elements of
R in the QR factorization of A.
'''
def qr_rank(A,tol=1e-15):
    m,n = A.shape
    assert m == n, "Input matrix must be square"
    Q,R = LA.qr(A)
    D = np.abs(np.diag(R))
    rank = np.sum(D > tol) 
    return rank, D



"""
DFO-LS
====

Derivative-Free Optimization for Least-Squares (DFO-LS) is a
nonlinear least-squares solver which only requires function values.

It solves the nonlinear least-squares problem:
    min_{x}  f(x) = r1(x)**2 + ... + rm(x)**2,
(optionally) subject to finitely many convex constraints,
where each function ri(x) is differentiable, possibly nonconvex.
Since the derivatives of ri(x) are never required or approximated,
the solver works when the evaluation of ri(x) is noisy.

----

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.


"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

# DFO-LS version
__version__ = '1.5.1'

# Main solver & exit flags
from .solver import *
__all__ = ['solve']




"""
Diagnostic Info
====

A class containing diagnostic information (optionally) produced by the solver.


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import pandas as pd
from .util import remove_scaling


__all__ = ['DiagnosticInfo', 'ITER_VERY_SUCCESSFUL', 'ITER_SUCCESSFUL', 'ITER_ACCEPTABLE_GEOM',
           'ITER_ACCEPTABLE_NO_GEOM', 'ITER_UNSUCCESSFUL_GEOM', 'ITER_UNSUCCESSFUL_NO_GEOM', 'ITER_SAFETY']


ITER_VERY_SUCCESSFUL = "Very successful"   # ratio >= 0.7, no geometry update
ITER_SUCCESSFUL = "Successful"  # 0.1 <= ratio < 0.7, no geometry update
ITER_ACCEPTABLE_GEOM = "Acceptable (geom fixed)"  # 0 <= ratio < 0.1, with geometry update
ITER_ACCEPTABLE_NO_GEOM = "Acceptable (geom not fixed)"  # 0 <= ratio < 0.1, without geometry update
ITER_UNSUCCESSFUL_GEOM = "Unsuccessful (geom fixed)"  # ratio < 0, with geometry update
ITER_UNSUCCESSFUL_NO_GEOM = "Unsuccessful (geom not fixed)"  # ratio < 0, without geometry update (possibly rho reduced)
ITER_SAFETY = "Safety"  # safety step taken (||s|| too small compared to rho)


class DiagnosticInfo(object):
    def __init__(self):
        self.data = {}
        # Initialise everything we want to store
        self.data["xk"] = []
        self.data["rk"] = []
        self.data["fk"] = []

        self.data["rho"] = []
        self.data["delta"] = []

        self.data["interpolation_error"] = []
        self.data["interpolation_condition_number"] = []
        self.data["interpolation_change_J_norm"] = []
        self.data["interpolation_total_residual"] = []
        self.data["poisedness"] = []
        self.data["max_distance_xk"] = []
        self.data["norm_gk"] = []
        self.data["norm_sk"] = []

        self.data["nruns"] = []
        self.data["nf"] = []
        self.data["nx"] = []
        self.data["npt"] = []
        self.data["nsamples"] = []
        self.data["iter_this_run"] = []
        self.data["iters_total"] = []

        self.data["iter_type"] = []
        self.data["ratio"] = []
        self.data["slow_iter"] = []
        return

    def to_dataframe(self, with_xk=False, with_rk=False):
        data_to_save = {}
        for key in self.data:
            if (key == "xk" and not with_xk) or (key == "rk" and not with_rk):
                continue  # skip
            data_to_save[key] = self.data[key]
        return pd.DataFrame(data_to_save)

    def to_csv(self, filename):
        df = self.to_dataframe()
        df.to_csv(filename)

    def save_info_from_control(self, control, nruns, iter_this_run, save_poisedness=True):
        self.data["iters_total"].append(len(self.data["iters_total"]))
        self.data["nruns"].append(nruns)
        self.data["iter_this_run"].append(iter_this_run)
        # Get what info we can out of a control
        self.data["nf"].append(control.nf)
        self.data["nx"].append(control.nx)
        self.data["delta"].append(control.delta)
        self.data["rho"].append(control.rho)
        # And from a model?
        self.data["npt"].append(control.model.npt())
        x, rvec, f, jac, nsamples = control.model.get_final_results()
        self.data["xk"].append(remove_scaling(x, control.scaling_changes))
        self.data["rk"].append(rvec)
        self.data["fk"].append(f)
        self.data["nsamples"].append(np.sum(control.model.nsamples))
        self.data["max_distance_xk"].append(np.sqrt(np.max(control.model.distances_to_xopt())))
        # Poisedness is expensive to compute (unlike everything else here), so allow the option to not calculate
        if save_poisedness:
            self.data["poisedness"].append(control.model.poisedness_constant(control.delta))
        else:
            self.data["poisedness"].append(0.0)
        # The other things we can't get just yet, so save a default value there for now
        self.data["interpolation_error"].append(None)
        self.data["interpolation_condition_number"].append(None)
        self.data["interpolation_change_J_norm"].append(None)
        self.data["interpolation_total_residual"].append(None)
        self.data["norm_gk"].append(None)
        self.data["norm_sk"].append(None)
        self.data["iter_type"].append(None)
        self.data["ratio"].append(None)
        self.data["slow_iter"].append(None)
        return

    def update_interpolation_information(self, interp_error, interp_cond_num, interp_total_resid, norm_change_J,
                                         norm_gk, norm_sk):
        self.data["interpolation_error"][-1] = interp_error
        self.data["interpolation_condition_number"][-1] = interp_cond_num
        self.data["interpolation_total_residual"][-1] = interp_total_resid
        self.data["interpolation_change_J_norm"][-1] = norm_change_J
        self.data["norm_gk"][-1] = norm_gk
        self.data["norm_sk"][-1] = norm_sk
        return

    def update_ratio(self, ratio):
        self.data["ratio"][-1] = ratio
        return

    def update_iter_type(self, iter_type):
        self.data["iter_type"][-1] = iter_type
        return

    def update_slow_iter(self, slow_iter):
        self.data["slow_iter"][-1] = slow_iter
        return




"""
Hessian Module
====

This module contains a class representing a Hessian matrix, by only storing
the upper triangular elements (since the matrix is symmetric).


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np


__all__ = ['Hessian']


class Hessian(object):
    def __init__(self, n, vals=None):
        self.n = n
        if vals is None:
            self.hq = np.zeros((n * (n + 1) // 2,), dtype=float)
        else:
            assert isinstance(vals, np.ndarray), "Can only set Hessian from NumPy array"
            assert len(vals.shape) in [1, 2], "Can only set Hessian from vector or matrix"
            if len(vals.shape) == 1:
                assert vals.shape[0] == self.n * (self.n + 1) // 2, "Incompatible n and input vector"
                self.hq = vals
            else:
                # Input was matrix
                assert vals.shape == (self.n, self.n), "Incompatible n and input matrix"
                self.hq = to_upper_triangular_vector(vals)

    def __len__(self):
        return len(self.hq)

    def dim(self):
        return self.n

    def shape(self):
        return self.hq.shape

    def upper_triangular(self):
        return self.hq

    def check_valid_index(self, i, j):
        assert i >= 0, "Index i must be >= 0"
        assert i <= self.n-1, "Index i must be <= n-1"
        assert j >= 0, "Index j must be >= 0"
        assert j <= self.n-1, "Index j must be <= n-1"

    def as_full(self):
        A = np.zeros((self.n, self.n))
        ih = -1
        for j in range(self.n):  # j = 0, ..., n-1
            for i in range(j + 1):  # i = 0, ..., j
                ih += 1
                A[i, j] = self.hq[ih]
                A[j, i] = self.hq[ih]
        return A

    def get_element(self, i, j):
        self.check_valid_index(i, j)
        ih = -1
        for k1 in range(self.n):
            for k2 in range(k1 + 1):
                ih += 1
                if (k1 == i and k2 == j) or (k1 == j and k2 == i):
                    return self.hq[ih]
        return None

    def set_element(self, i, j, val):
        self.check_valid_index(i, j)
        ih = -1
        for k1 in range(self.n):
            for k2 in range(k1 + 1):
                ih += 1
                if (k1 == i and k2 == j) or (k1 == j and k2 == i):
                    self.hq[ih] = val
        return

    def vec_mul(self, s):
        # Matrix-vector product
        assert isinstance(s, np.ndarray), "Can only multiply Hessian by a NumPy array"
        assert len(s.shape) == 1, "Can only multiply Hessian by a vector"
        assert s.shape == (self.n,), "Vector has incorrect length (expect %g, got %g)" % (self.n, s.shape[0])
        hs = np.zeros((self.n,))
        ih = -1
        for j in range(self.n):  # j = 0, ..., n-1
            for i in range(j + 1):  # i = 0, ..., j
                ih += 1
                if i < j:
                    hs[j] += self.hq[ih] * s[i]
                hs[i] += self.hq[ih] * s[j]
        return hs

    def __mul__(self, other):
        return self.vec_mul(other)


def to_upper_triangular_vector(A):
    n = A.shape[0]
    hq = np.zeros((n*(n+1)//2,))
    ih = -1
    for j in range(n):  # j = 0, ..., n-1
        for i in range(j + 1):  # i = 0, ..., j
            ih += 1
            hq[ih] = A[i,j]
    return hq



"""
Model
====

Maintain a class which represents an interpolating set, and its corresponding linear models
for each residual.
This class should calculate the various geometric quantities of interest to us.


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from math import sqrt
import numpy as np
import scipy.linalg as LA

from .trust_region import trsbox_geometry
from .util import sumsq, dykstra, remove_scaling

__all__ = ['Model']

module_logger = logging.getLogger(__name__) 


class Model(object):
    def __init__(self, npt, x0, r0, xl, xu, projections, r0_nsamples, h=None, argsh=(), n=None, m=None, abs_tol=1e-12, rel_tol=1e-20, precondition=True,
                 do_logging=True, scaling_changes=None):
        if n is None:
            n = len(x0)
        if m is None:
            m = len(r0)
        assert npt >= n + 1, "Require npt >= n+1 for linear models"
        assert x0.shape == (n,), "x0 has wrong shape (got %s, expect (%g,))" % (str(x0.shape), n)
        assert xl.shape == (n,), "xl has wrong shape (got %s, expect (%g,))" % (str(xl.shape), n)
        assert xu.shape == (n,), "xu has wrong shape (got %s, expect (%g,))" % (str(xu.shape), n)
        assert r0.shape == (m,), "r0 has wrong shape (got %s, expect (%g,))" % (str(r0.shape), m)
        self.do_logging = do_logging
        self.scaling_changes = scaling_changes
        self.dim = n
        self.resid_dim = m
        self.num_pts = npt
        self.npt_so_far = 1  # number of points added so far (with function values)

        self.h = h
        self.argsh = argsh

        # Initialise to blank some useful stuff
        # Interpolation points
        self.xbase = x0.copy()
        self.sl = xl - self.xbase  # lower bound w.r.t. xbase (require xpt >= sl)
        self.su = xu - self.xbase  # upper bound w.r.t. xbase (require xpt <= su)
        self.projections = projections
        self.points = np.zeros((npt, n))  # interpolation points w.r.t. xbase

        # Function values
        self.fval_v = np.inf * np.ones((npt, m))  # residuals for each xpt
        self.fval_v[0, :] = r0.copy()
        
        self.objval = np.inf * np.ones((npt, ))  # overall objective value for each xpt
        self.objval[0] = sumsq(r0)
        if h is not None:
            self.objval[0] += h(remove_scaling(x0, self.scaling_changes), *argsh)
        self.kopt = 0  # index of current iterate (should be best value so far)
        self.nsamples = np.zeros((npt,), dtype=int)  # number of samples used to evaluate objective at each point
        self.nsamples[0] = r0_nsamples
        self.objbeg = self.objval[0]  # f(x0), saved to check for sufficient reduction
        self.eval_num = np.zeros((npt,), dtype=int)  # which evaluation number (1-indexed, nx not nf) is currently stored self.points[k,:]
        self.eval_num[0] = 1

        # Termination criteria
        self.abs_tol = abs_tol
        self.rel_tol = rel_tol

        # Model information
        self.model_const = np.zeros((m, ))  # constant term for model m(s) = c + J*s
        self.model_jac = np.zeros((m, n))  # Jacobian term for model m(s) = c + J*s
        self.model_jac_eval_nums = None  # which evaluation numbers (1-indexed, nx not nf) were used to build model_jac

        # Saved point (in absolute coordinates) - always check this value before quitting solver
        self.xsave = None
        self.rsave = None
        self.objsave = None
        self.jacsave = None
        self.nsamples_save = None
        self.eval_num_save = None
        self.jacsave_eval_nums = None

        # Factorisation of interpolation matrix
        self.factorisation_current = False
        self.Q = None
        self.R = None
        self.qr_of_transpose = False  # is QR for W (finished growing) or W.T (growing)?
        self.precondition = precondition
        self.left_scaling = None  # preconditioning
        self.right_scaling = None

    def n(self):
        return self.dim

    def m(self):
        return self.resid_dim

    def npt(self):
        return min(self.num_pts, self.npt_so_far)

    def xopt(self, abs_coordinates=False):
        return self.xpt(self.kopt, abs_coordinates=abs_coordinates)

    def ropt(self):
        return self.fval_v[self.kopt, :]  # residuals for current iterate

    def objopt(self):
        return self.objval[self.kopt]

    def xpt(self, k, abs_coordinates=False):
        assert 0 <= k < self.npt(), "Invalid index %g" % k
        if not abs_coordinates:
            return np.minimum(np.maximum(self.sl, self.points[k, :].copy()), self.su)
        else:
            # Apply bounds and convert back to absolute coordinates
            if self.projections:
                return dykstra(self.projections, self.xbase + self.points[k,:])
            return self.xbase + np.minimum(np.maximum(self.sl, self.points[k, :]), self.su)

    def rvec(self, k):
        assert 0 <= k < self.npt(), "Invalid index %g" % k
        return self.fval_v[k, :]

    def objval(self, k):
        assert 0 <= k < self.npt(), "Invalid index %g" % k
        return self.objval[k]

    def as_absolute_coordinates(self, x, full_dykstra=False):
        # If x were an interpolation point, get the absolute coordinates of x
        if self.projections:
            return dykstra(self.projections, self.xbase + x)
        return self.xbase + np.minimum(np.maximum(self.sl, x), self.su)

    def xpt_directions(self, include_kopt=True):
        if include_kopt:
            ndirs = self.npt()
        else:
            ndirs = self.npt() - 1

        dirns = np.zeros((ndirs, self.n()))  # vector of directions xpt - xopt, excluding for xopt
        xopt = self.xopt()
        for k in range(self.npt()):
            if not include_kopt and k == self.kopt:
                continue  # skipt
            idx = k if include_kopt or k < self.kopt else k - 1
            dirns[idx, :] = self.xpt(k) - xopt
        return dirns

    def distances_to_xopt(self):
        sq_distances = np.zeros((self.npt(),))
        xopt = self.xopt()
        for k in range(self.npt()):
            sq_distances[k] = sumsq(self.points[k, :] - xopt)
        return sq_distances

    def change_point(self, k, x, rvec, eval_num, allow_kopt_update=True):
        # Update point k to x (w.r.t. xbase), with residual values fvec
        if k >= self.npt_so_far and self.npt_so_far < self.num_pts:
            assert k == self.npt_so_far, "Growing: updating wrong point"
            self.npt_so_far += 1
        else:
            assert 0 <= k < self.npt(), "Invalid index %g" % k

        self.points[k, :] = x.copy()
        self.fval_v[k, :] = rvec.copy()
        self.objval[k] = sumsq(rvec)
        if self.h is not None:
            self.objval[k] += self.h(remove_scaling(self.xbase + x, self.scaling_changes), *self.argsh)
        self.nsamples[k] = 1
        self.eval_num[k] = eval_num
        self.factorisation_current = False

        if allow_kopt_update and self.objval[k] < self.objopt():
            self.kopt = k
        return

    def swap_points(self, k1, k2):
        self.points[[k1, k2], :] = self.points[[k2, k1], :]
        self.fval_v[[k1, k2], :] = self.fval_v[[k2, k1], :]
        self.objval[[k1, k2]] = self.objval[[k2, k1]]
        self.eval_num[[k1, k2]] = self.eval_num[[k2, k1]]
        if self.kopt == k1:
            self.kopt = k2
        elif self.kopt == k2:
            self.kopt = k1
        self.factorisation_current = False
        return

    def add_new_sample(self, k, rvec_extra):
        # We have resampled at xpt(k) - add this information (fval and fval_v are averages of all samples)
        assert 0 <= k < self.npt(), "Invalid index %g" % k
        t = float(self.nsamples[k]) / float(self.nsamples[k] + 1)
        self.fval_v[k, :] = t * self.fval_v[k, :] + (1 - t) * rvec_extra
        # NOTE: how to sample when we have h? still at xpt(k), then add h(xpt(k)). Modify test if incorrect!
        self.objval[k] = sumsq(self.fval_v[k, :])
        if self.h is not None:
            self.objval[k] += self.h(remove_scaling(self.xbase + self.points[k, :], self.scaling_changes), *self.argsh)
        self.nsamples[k] += 1

        self.kopt = np.argmin(self.objval[:self.npt()])  # make sure kopt is always the best value we have
        return

    def add_new_point(self, x, rvec, eval_num):
        self.points = np.append(self.points, x.reshape((1, self.n())), axis=0)  # append row to xpt
        self.fval_v = np.append(self.fval_v, rvec.reshape((1, self.m())), axis=0)  # append row to fval_v
        obj = sumsq(rvec)
        if self.h is not None:
            obj += self.h(remove_scaling(self.xbase + x, self.scaling_changes), *self.argsh)
        self.objval = np.append(self.objval, obj)  # append entry to fval
        self.nsamples = np.append(self.nsamples, 1)  # add new sample number
        self.eval_num = np.append(self.eval_num, eval_num)  # add new evaluation number
        self.num_pts += 1  # make sure npt is updated
        self.npt_so_far += 1

        if obj < self.objopt():
            self.kopt = self.npt() - 1

        self.factorisation_current = False
        return

    def shift_base(self, xbase_shift):
        # Shifting xbase -> xbase + xbase_shift
        for k in range(self.npt()):
            self.points[k, :] = self.points[k, :] - xbase_shift
        self.xbase += xbase_shift
        self.sl = self.sl - xbase_shift
        self.su = self.su - xbase_shift
        self.factorisation_current = False

        # Update model
        self.model_const += np.dot(self.model_jac, xbase_shift)
        return

    def save_point(self, x, rvec, nsamples, eval_num, x_in_abs_coords=True):
        xabs = x.copy() if x_in_abs_coords else self.as_absolute_coordinates(x)
        obj = sumsq(rvec)
        if self.h is not None:
            obj += self.h(remove_scaling(xabs, self.scaling_changes), *self.argsh)
        if self.objsave is None or obj <= self.objsave:
            self.xsave = xabs
            self.rsave = rvec.copy()
            self.objsave = obj
            self.jacsave = self.model_jac.copy() if self.model_jac is not None else None
            self.nsamples_save = nsamples
            self.eval_num_save = eval_num
            self.jacsave_eval_nums = self.model_jac_eval_nums.copy() if self.model_jac_eval_nums is not None else None
            return True
        else:
            return False  # this value is worse than what we have already - didn't save

    def get_final_results(self):
        # Return x and objval for optimal point (either from xsave+objsave or kopt)
        if self.objsave is None or self.objopt() <= self.objsave:  # optimal has changed since xsave+objsave were last set
            return self.xopt(abs_coordinates=True).copy(), self.ropt().copy(), self.objopt(), self.model_jac.copy(), self.nsamples[self.kopt], self.eval_num[self.kopt], self.model_jac_eval_nums
        else:
            return self.xsave.copy(), self.rsave.copy(), self.objsave, self.jacsave, self.nsamples_save, self.eval_num_save, self.jacsave_eval_nums

    def min_objective_value(self):
        # Get termination criterion for f small: f <= abs_tol or f <= rel_tol * f0
        return max(self.abs_tol, self.rel_tol * self.objbeg)

    def model_value(self, d, d_based_at_xopt=True, with_const_term=False):
        if d_based_at_xopt:
            Jd = np.dot(self.model_jac, d + self.xopt())
        else:  # d based at xbase
            Jd = np.dot(self.model_jac, d)  # J * d
        return Jd + (self.model_const if with_const_term else 0.0)

    def interpolation_matrix(self):
        W = np.zeros((self.npt(), self.n()+1))
        if self.precondition:
            approx_delta = sqrt(np.max(self.distances_to_xopt()))  # largest distance to xopt ~ delta
        else:
            approx_delta = 1.0

        W[:, 0] = 1.0
        W[:, 1:] = self.xpt_directions(include_kopt=True) / approx_delta  # rows are (yt-xk)

        left_scaling = np.ones((self.npt(),))  # no left scaling
        right_scaling = np.ones((self.n() + 1,))
        right_scaling[1:] = 1.0 / approx_delta
        return W, left_scaling, right_scaling

    def factorise_geom_system(self):
        if not self.factorisation_current:
            W, self.left_scaling, self.right_scaling = self.interpolation_matrix()
            p, n = W.shape  # npt = p + 1
            if p >= n:  # finished growing (npt >= n+1)
                self.Q, self.R = LA.qr(W, mode='economic')  # reduced QR (saves memory)
                self.qr_of_transpose = False
            else:  # growing (npt < n+1)
                self.Q, self.R = LA.qr(W.T, mode='economic')  # reduced QR (saves memory)
                self.qr_of_transpose = True
            self.factorisation_current = True
        return

    def solve_geom_system(self, rhs):
        # To do preconditioning below, we will need to scale each column of A elementwise by the entries of some vector
        col_scale = lambda A, scale: (A.T*scale).T  # Uses the trick that A*x scales the 0th column of A by x[0], etc.

        if self.factorisation_current:
            if self.qr_of_transpose:
                # Growing case: solve underdetermined system W*x=rhs with W.T = Q*R
                # Golub & Van Loan (3rd edn), Algorithm 5.7.2
                Rb = LA.solve_triangular(self.R, col_scale(rhs, self.left_scaling), trans='T')  # R.T \ rhs
                return col_scale(np.dot(self.Q, Rb), self.right_scaling)  # minimal norm solution
            else:
                # Normal case: solve overdetermined system W*x=rhs with W=Q*R
                Qb = np.dot(self.Q.T, col_scale(rhs, self.left_scaling))
                return col_scale(LA.solve_triangular(self.R, Qb), self.right_scaling)
        else:
            if self.do_logging:
                module_logger.warning("model.solve_geom_system not using factorisation")
            W, left_scaling, right_scaling = self.interpolation_matrix()
            return col_scale(LA.lstsq(W, col_scale(rhs * left_scaling))[0], right_scaling)

    def interpolate_mini_models_svd(self, verbose=False, make_full_rank=False, min_sing_val=1e-6, sing_val_frac=1.0, max_jac_cond=1e8,
                                    get_chg_J=False, throw_error_on_nans=False):
        W, left_scaling, right_scaling = self.interpolation_matrix()
        self.factorise_geom_system()
        ls_interp_cond_num = np.linalg.cond(W) if verbose else 0.0  # scipy.linalg does not have condition number!

        # If not make_full_rank, Q is size (npt+n-1, npt+n), R is size (npt+n-1, n)
        # If make_full_rank, Q is size (2n, 2n), R is size (2n, n)
        xopt = self.xopt()
        ropt = self.ropt()
        fval_row_idx = np.arange(self.npt())  # indices of all rows
        norm_J_error = 0.0
        linalg_resid = 0.0

        if make_full_rank:
            # Remove old full-rank components of Jacobian
            Y = self.xpt_directions(include_kopt=False).T
            Qy, Ry = LA.qr(Y, mode='full')  # Qy is (n,n), Ry is (n,npt-1)=(n,p)
            Qhat = Qy[:, :Y.shape[1]]
            self.model_jac = np.dot(self.model_jac, np.dot(Qhat, Qhat.T))

        rhs = self.fval_v[fval_row_idx, :]  # size npt * m
        if np.any(np.isnan(rhs)) and throw_error_on_nans:
            if self.do_logging:
                module_logger.warning("model.interpolate_mini_models_svd: NaNs encountered in objective evaluations, raising error")
            raise np.linalg.LinAlgError("NaN encountered in objective evaluations")
        try:
            dg = self.solve_geom_system(rhs)  # size (n+1)*m
        except LA.LinAlgError:
            return False, None, None, None, None  # flag error
        except ValueError:
            return False, None, None, None, None  # flag error (e.g. inf or NaN encountered)
        if not np.all(np.isfinite(dg)):  # another check for inf or NaN
            return False, None, None, None, None
        J_old = self.model_jac.copy()
        self.model_jac = dg[1:,:].T
        self.model_const = dg[0,:] - np.dot(self.model_jac, xopt)  # shift base to xbase
        self.model_jac_eval_nums = self.eval_num.copy()
        if verbose or get_chg_J:
            norm_J_error = np.linalg.norm(self.model_jac - J_old, ord='fro')**2
            linalg_resid = np.linalg.norm(W.dot(dg) - rhs)**2

        if make_full_rank:
            try:
                U, s, Vt = LA.svd(self.model_jac, full_matrices=False)  # U is (m,k), s has length k, Vt is (k,n), where k=min(m,n)
            except LA.LinAlgError:
                return False, None, None, None, None  # flag error
            k = min(self.n(), self.m())
            r = min(self.npt_so_far - 1, self.n(), self.m())  # current number of directions (i.e. rank of J)
            floor_val = max(s[0]/max_jac_cond, sing_val_frac * s[r-1], min_sing_val)
            s = np.maximum(s, floor_val)
            S = LA.diagsvd(s, k, k)  # s from vector to matrix of correct shape
            self.model_jac = np.dot(U, np.dot(S, Vt))  # reconstruct J from new svd

        interp_error = 0.0
        if verbose:
            for k in range(self.npt()):
                r_pred = self.model_value(self.xpt(k), d_based_at_xopt=False, with_const_term=True)
                interp_error += self.nsamples[k] * sumsq(self.fval_v[k, :] - r_pred)

        return True, interp_error, sqrt(norm_J_error), linalg_resid, ls_interp_cond_num  # flag ok

    def build_full_model(self):
        # Build full least squares model from mini-models
        # Centred around xopt
        r = self.model_const + np.dot(self.model_jac, self.xopt())  # constant term (for inexact interpolation)
        J = self.model_jac

        # Apply scaling based on convention for objective - this code uses sumsq(rvec) not 0.5*sumsq(rvec)
        g = 2.0 * np.dot(J.T, r)  # n-vector
        H = 2.0 * np.dot(J.T, J)
        return g, H

    def lagrange_gradient(self, k=None, factorise_first=True):
        if factorise_first:
            self.factorise_geom_system()

        if k is not None:
            assert 0 <= k < self.npt(), "Invalid index %g" % k
            rhs = np.zeros((self.npt(),))
            rhs[k] = 1.0
        else:
            rhs = np.eye(self.npt())  # find all Lagrange polynomials
        soln = self.solve_geom_system(rhs)

        if k is not None:
            c = soln[0]
            g = soln[1:]
            return c, g  # constant, gradient [all based at xopt]
        else:
            cs = soln[0, :]
            gs = soln[1:, :]
            return cs, gs  # constant terms in each entry and gradient terms in each col [all based at xopt]

    def poisedness_constant(self, delta, xbase=None, xbase_in_abs_coords=True):
        # Calculate the poisedness constant of the current interpolation set in B(xbase, delta)
        # if xbase is None, use self.xopt()
        overall_max = None
        if xbase is None:
            xbase = self.xopt()
        elif xbase_in_abs_coords:
            xbase = xbase - self.xbase  # shift to correct position
        # Calculate all Lagrange polynomials at once
        self.factorise_geom_system()
        rhs = np.eye(self.npt())  # values to interpolate
        soln = self.solve_geom_system(rhs)
        for k in range(self.npt()):
            # Extract Lagrange poly from soln matrix (based at xopt)
            c = soln[0,k]; g = soln[1:, k]
            newc = c + np.dot(g, xbase - self.xopt())  # based at xbase
            # Solve problem: bounds are sl <= x <= su, and ||x-xopt|| <= delta
            xmax = trsbox_geometry(xbase, newc, g, self.sl, self.su, delta)
            lmax = abs(c + np.dot(g, xmax - self.xopt()))  # evaluate Lagrange poly
            if overall_max is None or lmax > overall_max:
                overall_max = lmax
        return overall_max




"""
Controller
====

A class containing all the actions we want to perform during the algorithm.
This allows the code for the actual solver to focus on the logic & flow of the
algorithm, not the specific implementation details.


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

The development of this software was sponsored by NAG Ltd. (http://www.nag.co.uk)
and the EPSRC Centre For Doctoral Training in Industrially Focused Mathematical
Modelling (EP/L015803/1) at the University of Oxford. Please contact NAG for
alternative licensing.

"""

# Ensure compatibility with Python 2
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from math import log, sqrt
import numpy as np
import scipy.linalg as LA

from .model import *
from .trust_region import *
from .util import *

__all__ = ['Controller', 'ExitInformation', 'EXIT_SLOW_WARNING', 'EXIT_MAXFUN_WARNING', 'EXIT_SUCCESS',
           'EXIT_INPUT_ERROR', 'EXIT_TR_INCREASE_ERROR', 'EXIT_LINALG_ERROR', 'EXIT_FALSE_SUCCESS_WARNING',
           'EXIT_AUTO_DETECT_RESTART_WARNING', 'EXIT_EVAL_ERROR']

module_logger = logging.getLogger(__name__) 

EXIT_TR_INCREASE_WARNING = 5  # warning, TR increase in proj constrained case - likely due to multiple active constraints
EXIT_AUTO_DETECT_RESTART_WARNING = 4  # warning, auto-detected restart criteria
EXIT_FALSE_SUCCESS_WARNING = 3  # warning, maximum fake successful steps reached
EXIT_SLOW_WARNING = 2  # warning, maximum number of slow (successful) iterations reached
EXIT_MAXFUN_WARNING = 1  # warning, reached max function evals
EXIT_SUCCESS = 0  # successful finish (rho=rhoend, sufficient objective reduction, or everything in noise level)
EXIT_INPUT_ERROR = -1  # error, bad inputs
EXIT_TR_INCREASE_ERROR = -2  # error, trust region step increased model value
EXIT_LINALG_ERROR = -3  # error, linalg error (singular matrix encountered)
EXIT_EVAL_ERROR = -4  # error, objective evaluation error (e.g. nan result received)


class ExitInformation(object):
    def __init__(self, flag, msg_details):
        self.flag = flag
        self.msg = msg_details

    def flag(self):
        return self.flag

    def message(self, with_stem=True):
        if not with_stem:
            return self.msg
        elif self.flag == EXIT_SUCCESS:
            return "Success: " + self.msg
        elif self.flag == EXIT_SLOW_WARNING:
            return "Warning (slow progress): " + self.msg
        elif self.flag == EXIT_MAXFUN_WARNING:
            return "Warning (max evals): " + self.msg
        elif self.flag == EXIT_TR_INCREASE_WARNING:
            return "Warning (trust region increase): " + self.msg
        elif self.flag == EXIT_INPUT_ERROR:
            return "Error (bad input): " + self.msg
        elif self.flag == EXIT_TR_INCREASE_ERROR:
            return "Error (trust region increase): " + self.msg
        elif self.flag == EXIT_LINALG_ERROR:
            return "Error (linear algebra): " + self.msg
        elif self.flag == EXIT_FALSE_SUCCESS_WARNING:
            return "Warning (max false good steps): " + self.msg
        elif self.flag == EXIT_EVAL_ERROR:
            return "Error (function evaluation): " + self.msg
        else:
            return "Unknown exit flag: " + self.msg

    def able_to_do_restart(self):
        if self.flag in [EXIT_TR_INCREASE_ERROR, EXIT_TR_INCREASE_WARNING, EXIT_LINALG_ERROR, EXIT_SLOW_WARNING, EXIT_AUTO_DETECT_RESTART_WARNING, EXIT_EVAL_ERROR]:
            return True
        elif self.flag in [EXIT_MAXFUN_WARNING, EXIT_INPUT_ERROR]:
            return False
        else:
            # Successful step (rho=rhoend, noise level termination, or value small)
            return "sufficiently small" not in self.msg  # restart for rho=rhoend and noise level termination


class Controller(object):
    def __init__(self, objfun, argsf, x0, r0, r0_nsamples, xl, xu, projections, npt, rhobeg, rhoend, nf, nx, maxfun, params,
                 scaling_changes, do_logging, h=None, lh=None, argsh = (), prox_uh=None, argsprox = ()):
        self.do_logging = do_logging
        self.objfun = objfun
        self.h = h
        self.argsf = argsf
        self.argsh = argsh
        self.lh = lh
        self.prox_uh = prox_uh #TODO: add instruction for prox_uh
        self.argsprox = argsprox
        self.maxfun = maxfun
        self.model = Model(npt, x0, r0, xl, xu, projections, r0_nsamples, h=self.h, argsh = argsh, precondition=params("interpolation.precondition"),
                           abs_tol = params("model.abs_tol"), rel_tol = params("model.rel_tol"), do_logging=do_logging, scaling_changes=scaling_changes)
        self.nf = nf
        self.nx = nx
        self.rhobeg = rhobeg
        self.delta = rhobeg
        self.rho = rhobeg
        self.rhoend = rhoend
        self.diffs = [0.0, 0.0, 0.0]
        self.finished_growing = False
        self.finished_halfway_growing = False
        # For measuing slow iterations
        self.last_iters_step_taken = []
        self.last_fopts_step_taken = []
        self.num_slow_iters = 0  # how many consecutive slow iterations have we had so far
        # For determining when to reduce rho
        self.last_successful_iter = 0  # when ||d|| >= rho
        # For counting the number of soft restarts
        self.last_successful_run = 0
        self.last_run_fopt = sumsq(r0)
        self.scaling_changes = scaling_changes

    def n(self):
        return self.model.n()

    def m(self):
        return self.model.m()

    def npt(self):
        return self.model.npt()

    def initialise_coordinate_directions(self, number_of_samples, num_directions, params):
        if self.do_logging:
            module_logger.debug("Initialising with coordinate directions")
        # self.model already has x0 evaluated, so only need to initialise the other points
        # num_directions = params("growing.ndirs_initial")
        assert self.model.num_pts <= (self.n() + 1) * (self.n() + 2) // 2, "prelim: must have npt <= (n+1)(n+2)/2"
        assert 1 <= num_directions < self.model.num_pts, "Initialisation: must have 1 <= ndirs_initial < npt"


        if self.model.projections:
            D = np.zeros((self.n(),self.n()))
            k = 0
            while k < self.n():
                ek = np.zeros(self.n())
                ek[k] = 1
                p = np.dot(ek,min(1,self.delta))
                yk = dykstra(self.model.projections, self.model.xbase + p, max_iter=params("dykstra.max_iters"), tol=params("dykstra.d_tol"))
                D[k,:] = yk - self.model.xbase

                k += 1 # move on to next point

            # Have at least one L.D. vector, try negative direction on bad one first
            k = 0
            mr_tol = params("matrix_rank.r_tol")
            D_rank, diag = qr_rank(D,tol=mr_tol)
            while D_rank != num_directions and k < self.n():
                if diag[k] < mr_tol:
                    ek = np.zeros(self.n())
                    ek[k] = 1
                    p = -np.dot(ek,min(1,self.delta))
                    yk = dykstra(self.model.projections, self.model.xbase + p, max_iter=params("dykstra.max_iters"), tol=params("dykstra.d_tol"))
                    dk = D[k,:].copy()
                    D[k,:] = yk - self.model.xbase
                    D_rank2, _diag2 = qr_rank(D,tol=params("matrix_rank.r_tol"))
                    if D_rank2 <= D_rank:
                        # Did not improve rank, revert change
                        D[k,:] = dk
                    # rank was improved, update D_rank for next comparison
                    D_rank = D_rank2
                k += 1

            # Try random combination of negatives...
            k = 0
            slctr = np.random.randint(0, 1+1, self.n()) # generate rand binary "selector" array
            D_rank, diag = qr_rank(D,tol=params("matrix_rank.r_tol"))
            while D_rank != num_directions and k < 100*self.n():
                if slctr[k%self.n()] == 1: # if selector says make -ve, make -ve
                    ek = np.zeros(self.n())
                    ek[k%self.n()] = 1
                    p = -np.dot(ek,min(1,self.delta))
                    yk = dykstra(self.model.projections, self.model.xbase + p, max_iter=params("dykstra.max_iters"), tol=params("dykstra.d_tol"))
                    dk = D[k%self.n(),:].copy()
                    D[k%self.n(),:] = yk - self.model.xbase
                    D_rank2, _diag2 = qr_rank(D,tol=params("matrix_rank.r_tol"))
                    if D_rank2 <= D_rank:
                        # Did not improve rank, revert change
                        D[k%self.n(),:] = dk
                    # rank was improved, update D_rank for next comparison
                    D_rank = D_rank2

                # Go again
                slctr = np.random.randint(0, 1+1, self.n())
                k += 1

            # Set still not L.I? Try random directions
            i = 0
            D_rank, diag = qr_rank(D,tol=params("matrix_rank.r_tol"))
            while D_rank != num_directions and i <= 100*num_directions:
                k = 0
                while k < self.n():
                    if diag[k] < mr_tol:
                        p = np.random.normal(size=self.n())
                        p = p/np.linalg.norm(p)
                        p = np.dot(p,min(1,self.delta))
                        yk = dykstra(self.model.projections, self.model.xbase + p, max_iter=params("dykstra.max_iters"), tol=params("dykstra.d_tol"))
                        dk = D[k,:].copy()
                        D[k,:] = yk - self.model.xbase
                        D_rank2, _diag2 = qr_rank(D,tol=params("matrix_rank.r_tol"))
                        if D_rank2 <= D_rank:
                            # Did not improve rank, revert change
                            D[k,:] = dk
                        # rank was improved, update D_rank for next comparison
                        D_rank = D_rank2
                    k += 1
                i += 1

            if D_rank != num_directions:
                raise RuntimeError("Unable to generate suitable initial directions")

            # we have a L.I set of interpolation points
            for k in range(0,self.n()):
                # Evaluate objective at this new point
                x = self.model.as_absolute_coordinates(D[k, :])
                rvec_list, obj_list, num_samples_run, exit_info = self.evaluate_objective(x, number_of_samples, params)

                # Handle exit conditions (f < min obj value or maxfun reached)
                if exit_info is not None:
                    if num_samples_run > 0:
                        self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                              x_in_abs_coords=True)
                    return exit_info  # return & quit

                # Otherwise, add new results (increments model.npt_so_far)
                self.model.change_point(k+1, x - self.model.xbase, rvec_list[0, :], self.nx)  # expect step, not absolute x
                for i in range(1, num_samples_run):
                    self.model.add_new_sample(k+1, rvec_extra=rvec_list[i, :])
            
            return None   # return & continue

        at_lower_boundary = (self.model.sl > -0.01 * self.delta)  # sl = xl - x0, should be -ve, actually < -rhobeg
        at_upper_boundary = (self.model.su < 0.01 * self.delta)  # su = xu - x0, should be +ve, actually > rhobeg
        
        if params("init.run_in_parallel") and num_directions <= self.n():
            # Can do all the evaluation in parallel if <= n+1 interpolation points, but if larger
            # then the step depends on the function value at previous steps and does point swapping
            xpts_added = np.zeros((num_directions + 1, self.n()))
            eval_obj_results = []
            for k in range(1, num_directions + 1):  # k = 1, ..., num_directions
                # always have k = 1, ..., n since num_directions <= n
                dirn = k - 1  # direction to move in (0,...,n-1)
                stepa = self.delta if not at_upper_boundary[dirn] else -self.delta # take a +delta step if at lower, -delta if at upper
                stepb = None
                xpts_added[k, dirn] = stepa # set new (relative) point to the step since we haven't done any moving, so relative point is all zeros.
                
                # Evaluate objective at this new point
                x = self.model.as_absolute_coordinates(xpts_added[k, :])
                eval_obj_results.append(self.evaluate_objective(x, number_of_samples, params))
            
            # Evaluations done, now add to the model
            for k in range(1, num_directions + 1):
                x = self.model.as_absolute_coordinates(xpts_added[k, :])
                rvec_list, obj_list, num_samples_run, exit_info = eval_obj_results[k-1]
                # Handle exit conditions (f < min obj value or maxfun reached)
                if exit_info is not None:
                    if num_samples_run > 0:
                        self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                              x_in_abs_coords=True)
                    return exit_info  # return & quit

                # Otherwise, add new results (increments model.npt_so_far)
                self.model.change_point(k, x - self.model.xbase, rvec_list[0, :], self.nx)  # expect step, not absolute x
                for i in range(1, num_samples_run):
                    self.model.add_new_sample(k, rvec_extra=rvec_list[i, :])
        else:
            xpts_added = np.zeros((num_directions + 1, self.n()))
            for k in range(1, num_directions + 1):
                # k = 0 --> base point (xpt = 0)  [ not here]
                # k = 1, ..., 2n --> coordinate directions [1,...,n and n+1,...,2n]
                # k = 2n+1, ..., (n+1)(n+2)/2 --> off-diagonal directions
                if 1 <= k < self.n() + 1:  # first step along coord directions
                    dirn = k - 1  # direction to move in (0,...,n-1)
                    stepa = self.delta if not at_upper_boundary[dirn] else -self.delta # take a +delta step if at lower, -delta if at upper
                    stepb = None
                    xpts_added[k, dirn] = stepa # set new (relative) point to the step since we haven't done any moving, so relative point is all zeros.

                elif self.n() + 1 <= k < 2 * self.n() + 1:  # second step along coord directions
                    dirn = k - self.n() - 1  # direction to move in (0,...,n-1)
                    stepa = xpts_added[k - self.n(), dirn] # previous step
                    stepb = -self.delta # new step
                    if at_lower_boundary[dirn]:
                        # if at lower boundary, set the second step to be +ve
                        stepb = min(2.0 * self.delta, self.model.su[dirn])  # su = xu - x0, should be +ve
                    if at_upper_boundary[dirn]:
                        # if at upper boundary, set the second step to be -ve
                        stepb = max(-2.0 * self.delta, self.model.sl[dirn])  # sl = xl - x0, should be -ve
                    xpts_added[k, dirn] = stepb

                else:  # k = 2n+1, ..., (n+1)(n+2)/2
                    # p = (k - 1) % n + 1  # cycles through (1,...,n), starting at 2n+1 --> 1
                    # l = (k - 2 * n - 1) / n + 1  # (1,...,1, 2, ..., 2, etc.) where each number appears n times
                    # q = (p + l if p + l <= n else p + l - n)
                    stepa = None
                    stepb = None
                    itemp = (k - self.n() - 1) // self.n()
                    q = k - itemp * self.n() - self.n()
                    p = q + itemp
                    if p > self.n():
                        p, q = q, p - self.n()  # does swap correctly in Python

                    xpts_added[k, p - 1] = xpts_added[p, p - 1]
                    xpts_added[k, q - 1] = xpts_added[q, q - 1]

                # Evaluate objective at this new point
                x = self.model.as_absolute_coordinates(xpts_added[k, :])
                rvec_list, obj_list, num_samples_run, exit_info = self.evaluate_objective(x, number_of_samples, params)

                # Handle exit conditions (f < min obj value or maxfun reached)
                if exit_info is not None:
                    if num_samples_run > 0:
                        self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                              x_in_abs_coords=True)
                    return exit_info  # return & quit

                # Otherwise, add new results (increments model.npt_so_far)
                self.model.change_point(k, x - self.model.xbase, rvec_list[0, :], self.nx)  # expect step, not absolute x
                for i in range(1, num_samples_run):
                    self.model.add_new_sample(k, rvec_extra=rvec_list[i, :])

                # If k exceeds N+1, then the positions of the k-th and (k-N)-th interpolation
                # points may be switched, in order that the function value at the first of them
                # contributes to the off-diagonal second derivative terms of the initial quadratic model.
                # Note: this works because the steps for (k) and (k-n) points were in the same coordinate direction
                if self.n() + 1 <= k < 2 * self.n() + 1:
                    # Only swap if steps were in different directions AND new pt has lower objective
                    if stepa * stepb < 0.0 and self.model.objval[k] < self.model.objval[k - self.n()]:
                        xpts_added[[k, k-self.n()]] = xpts_added[[k-self.n(), k]]

        return None   # return & continue

    def initialise_random_directions(self, number_of_samples, num_directions, params):
        if self.do_logging:
            module_logger.debug("Initialising with random orthogonal directions")
        # self.model already has x0 evaluated, so only need to initialise the other points
        assert 1 <= num_directions < self.model.num_pts, "Initialisation: must have 1 <= ndirs_initial < npt"

        # Get ndirs_initial random orthogonal directions
        xopt = self.model.xopt()
        if params("init.random_directions_make_orthogonal"):
            dirns = random_orthog_directions_within_bounds(num_directions, self.delta, self.model.sl - xopt,
                                                           self.model.su - xopt)
        else:
            dirns = random_directions_within_bounds(num_directions, self.delta, self.model.sl - xopt,
                                                    self.model.su - xopt)
        # Now add the random directions
        if params("init.run_in_parallel"):
            # Run all results in parallel first, then process
            eval_obj_results = []
            for ndirns in range(num_directions):
                new_point = xopt + dirns[ndirns, :]  # alway base move around best value so far

                # Evaluate objective
                x = self.model.as_absolute_coordinates(new_point)
                eval_obj_results.append(self.evaluate_objective(x, number_of_samples, params))

            for ndirns in range(num_directions):
                new_point = xopt + dirns[ndirns, :]  # alway base move around best value so far
                x = self.model.as_absolute_coordinates(new_point)
                rvec_list, obj_list, num_samples_run, exit_info = eval_obj_results[ndirns]
                # Handle exit conditions (f < min obj value or maxfun reached)
                if exit_info is not None:
                    if num_samples_run > 0:
                        self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                              x_in_abs_coords=True)
                    return exit_info  # return & quit

                # Otherwise, add new results (increments model.npt_so_far)
                self.model.change_point(1 + ndirns, x - self.model.xbase,
                                        rvec_list[0, :], self.nx)  # expect step, not absolute x
                for i in range(1, num_samples_run):
                    self.model.add_new_sample(1 + ndirns, rvec_extra=rvec_list[i, :])
        else:
            for ndirns in range(num_directions):
                new_point = xopt + dirns[ndirns, :]  # alway base move around best value so far

                # Evaluate objective
                x = self.model.as_absolute_coordinates(new_point)
                rvec_list, obj_list, num_samples_run, exit_info = self.evaluate_objective(x, number_of_samples, params)

                # Handle exit conditions (f < min obj value or maxfun reached)
                if exit_info is not None:
                    if num_samples_run > 0:
                        self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                              x_in_abs_coords=True)
                    return exit_info  # return & quit

                # Otherwise, add new results (increments model.npt_so_far)
                self.model.change_point(1 + ndirns, x - self.model.xbase, rvec_list[0, :], self.nx)  # expect step, not absolute x
                for i in range(1, num_samples_run):
                    self.model.add_new_sample(1 + ndirns, rvec_extra=rvec_list[i, :])

        return None

    def add_new_direction_while_growing(self, number_of_samples, params, min_num_steps=0):
        num_steps = max(params('growing.num_new_dirns_each_iter'), min_num_steps)
        step_length = params('growing.delta_scale_new_dirns') * self.delta
        if num_steps < 1:  # not actually adding new directions
            return None

        # Step from xopt along a random direction orthogonal to other yt (or multiple mutually orthogonal steps)
        xopt = self.model.xopt()
        dirns = random_directions_within_bounds(num_steps, step_length, self.model.sl - xopt, self.model.su - xopt)
        # Make direction orthogonal
        Y = self.model.xpt_directions(include_kopt=False).T  # columns are the current set of directions
        Q, R = LA.qr(Y, mode='economic')  # columns of Q are orthonormal basis for current set of directions
        for k in range(Q.shape[1]):
            qk = Q[:, k]
            for j in range(dirns.shape[0]):
                dirns[j, :] = dirns[j, :] - np.dot(dirns[j, :], qk) * qk

        # Evaluate the points
        for j in range(num_steps):
            xnew = self.model.xopt() + (step_length / LA.norm(dirns[j, :])) * dirns[j, :]
            x = self.model.as_absolute_coordinates(xnew)
            rvec_list, obj_list, num_samples_run, exit_info = self.evaluate_objective(x, number_of_samples, params)

            # Handle exit conditions (f < min obj value or maxfun reached)
            if exit_info is not None:
                if num_samples_run > 0:
                    self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                          x_in_abs_coords=True)
                return exit_info  # return & quit

            if self.model.npt() < self.model.num_pts:  # still growing
                kmin = self.model.npt()
            else:  # full set - choose in usual way
                kmin, linalg_error = self.choose_point_to_replace(xnew - self.model.xopt(), skip_kopt=True)

                if linalg_error:
                    exit_info = ExitInformation(EXIT_LINALG_ERROR, "Singular matrix when finding kmin (in main loop)")
                    return exit_info  # return & quit

            # Otherwise, add new results
            self.model.change_point(kmin, xnew, rvec_list[0, :], self.nx)  # expect step, not absolute x
            for i in range(1, num_samples_run):
                self.model.add_new_sample(kmin, rvec_extra=rvec_list[i, :])

        return None

    def get_new_direction_for_growing(self, step_length):
        # Step from xopt along a random direction orthogonal to other yt (or multiple mutually orthogonal steps)
        xopt = self.model.xopt()
        dirn = random_directions_within_bounds(1, step_length, self.model.sl - xopt, self.model.su - xopt)[0, :]
        # Make direction orthogonal
        Y = self.model.xpt_directions(include_kopt=False).T  # columns are the current set of directions
        Q, R = LA.qr(Y, mode='economic')  # columns of Q are orthonormal basis for current set of directions
        for k in range(Q.shape[1]):
            qk = Q[:, k]
            dirn = dirn - np.dot(dirn, qk) * qk

        return dirn * (step_length / LA.norm(dirn))

    def evaluate_criticality_measure(self, params):
        # Calculate criticality measure for regularized problems (h is not None)
        
        # Build model for full least squares function
        gopt, H = self.model.build_full_model()
        
        if np.any(np.isnan(gopt)) or np.any(np.isnan(H)) or not np.all(np.isfinite(gopt)) or not np.all(np.isfinite(H)):
            module_logger.debug("nan/inf values in gopt and/or H, skipping ctrsbox_sfista (criticality measure calc)")
            # d = np.zeros(gopt.shape)
            # gnew = gopt.copy()
            # crvmin = -1
            return np.inf
        
        # NOTE: smaller params here to get more iterations in S-FISTA
        func_tol = params("func_tol.criticality_measure") * self.delta
        if self.model.projections:
            d, gnew, crvmin = ctrsbox_sfista(self.model.xopt(abs_coordinates=True), gopt, np.zeros(H.shape), self.model.projections, 1,
                                self.h, self.lh, self.prox_uh, argsh = self.argsh, argsprox=self.argsprox, func_tol=func_tol, 
                                max_iters=params("func_tol.max_iters"), d_max_iters=params("dykstra.max_iters"), d_tol=params("dykstra.d_tol"),
                                scaling_changes=self.scaling_changes, sfista_iters_scale=params("sfista.max_iters_scaling"))
        else:
            proj = lambda x: pbox(x, self.model.sl, self.model.su)
            d, gnew, crvmin = ctrsbox_sfista(self.model.xopt(abs_coordinates=True), gopt, np.zeros(H.shape), [proj], 1,
                                self.h, self.lh, self.prox_uh, argsh = self.argsh, argsprox=self.argsprox, func_tol=func_tol, 
                                max_iters=params("func_tol.max_iters"), d_max_iters=params("dykstra.max_iters"), d_tol=params("dykstra.d_tol"),
                                scaling_changes=self.scaling_changes, sfista_iters_scale=params("sfista.max_iters_scaling"))
        
        # Calculate criticality measure
        criticality_measure = self.h(remove_scaling(self.model.xopt(abs_coordinates=True), self.scaling_changes), *self.argsh) - model_value(gopt, np.zeros(H.shape), d, self.model.xopt(abs_coordinates=True), self.h, self.argsh, self.scaling_changes)
        return criticality_measure

    def trust_region_step(self, params, criticality_measure=1e-2):
        # Build model for full least squares function
        gopt, H = self.model.build_full_model()
        # Build func_tol for trust region step
        # QUESTION: c1 = min{1, 1/delta_max^2}, but choose c1=1here; choose maxhessian = max(||H||_2,1)
        # QUESTION: when criticality_measure = 0? choose max(criticality_measure,1)
        func_tol = (1-params("func_tol.tr_step")) * 1 * max(criticality_measure,1) * min(self.delta, max(criticality_measure,1) / max(np.linalg.norm(H, 2),1))

        if self.h is None:
            if self.model.projections:
                # Running PGD/SFISTA is generally slower than trsbox, so don't do this if gopt or H have bad values
                # (this will ultimately lead to a manual setting of d=0 and calling a safety step anyway)
                if np.any(np.isnan(gopt)) or np.any(np.isnan(H)) or not np.all(np.isfinite(gopt)) or not np.all(np.isfinite(H)):
                    module_logger.debug("nan/inf values in gopt and/or H, skipping ctrsbox_pgd")
                    d = np.zeros(gopt.shape)
                    gnew = gopt.copy()
                    crvmin = -1
                else:
                    d, gnew, crvmin = ctrsbox_pgd(self.model.xopt(abs_coordinates=True), gopt, H, self.model.projections, self.delta, d_max_iters=params("dykstra.max_iters"), d_tol=params("dykstra.d_tol"))
            else:
                d, gnew, crvmin = trsbox(self.model.xopt(), gopt, H, self.model.sl, self.model.su, self.delta)
        else:
            # Running PGD/SFISTA is generally slower than trsbox, so don't do this if gopt or H have bad values
            # (this will ultimately lead to a manual setting of d=0 and calling a safety step anyway)
            if np.any(np.isnan(gopt)) or np.any(np.isnan(H)) or not np.all(np.isfinite(gopt)) or not np.all(np.isfinite(H)):
                module_logger.debug("nan/inf values in gopt and/or H, skipping ctrsbox_sfista")
                d = np.zeros(gopt.shape)
                gnew = gopt.copy()
                crvmin = -1
            elif self.model.projections:
                d, gnew, crvmin = ctrsbox_sfista(self.model.xopt(abs_coordinates=True), gopt, H, self.model.projections, self.delta,
                                     self.h, self.lh, self.prox_uh, argsh = self.argsh, argsprox=self.argsprox, func_tol=func_tol, 
                                     max_iters=params("func_tol.max_iters"), d_max_iters=params("dykstra.max_iters"), d_tol=params("dykstra.d_tol"),
                                     scaling_changes=self.scaling_changes, sfista_iters_scale=params("sfista.max_iters_scaling"))
            else:
                # NOTE: alternative way if using trsbox
                # d, gnew, crvmin = trsbox(self.model.xopt(), gopt, H, self.model.sl, self.model.su, self.delta)
                proj = lambda x: pbox(x, self.model.sl, self.model.su)
                d, gnew, crvmin = ctrsbox_sfista(self.model.xopt(abs_coordinates=True), gopt, H, [proj], self.delta,
                                      self.h, self.lh, self.prox_uh, argsh = self.argsh, argsprox=self.argsprox, func_tol=func_tol,
                                      max_iters=params("func_tol.max_iters"), d_max_iters=params("dykstra.max_iters"), d_tol=params("dykstra.d_tol"),
                                      scaling_changes=self.scaling_changes, sfista_iters_scale=params("sfista.max_iters_scaling"))
            
            # NOTE: check sufficient decrease. If increase in the model, set zero step
            pred_reduction = self.h(remove_scaling(self.model.xopt(abs_coordinates=True), self.scaling_changes), *self.argsh) - model_value(gopt, H, d, self.model.xopt(abs_coordinates=True), self.h, self.argsh, self.scaling_changes)
            if pred_reduction < 0.0:
                d = np.zeros(d.shape)
        
        return d, gopt, H, gnew, crvmin

    def geometry_step(self, knew, adelt, number_of_samples, params):
        if self.do_logging:
            module_logger.debug("Running geometry-fixing step")
        try:
            c, g = self.model.lagrange_gradient(knew)
            # c = 1.0 if knew == self.model.kopt else 0.0  # based at xopt, just like d
            if self.model.projections:
                # Solve problem: use projection onto arbitrary constraints, and ||xnew-xopt|| <= adelt
                step = ctrsbox_geometry(self.model.xopt(abs_coordinates=True), c, g, self.model.projections, adelt, d_max_iters=params("dykstra.max_iters"), d_tol=params("dykstra.d_tol"))
                xnew = self.model.xopt() + step
            else:
                # Solve problem: bounds are sl <= xnew <= su, and ||xnew-xopt|| <= adelt
                xnew = trsbox_geometry(self.model.xopt(), c, g, np.minimum(self.model.sl, 0.0), np.maximum(self.model.su, 0.0), adelt)
        except LA.LinAlgError:
            exit_info = ExitInformation(EXIT_LINALG_ERROR, "Singular matrix encountered in geometry step")
            return exit_info  # didn't fix geometry - return & quit

        gopt, H = self.model.build_full_model()  # save here, to calculate predicted value from geometry step
        objopt = self.model.objopt()  # again, evaluate now, before model.change_point()
        d = xnew - self.model.xopt()
        x = self.model.as_absolute_coordinates(xnew)
        rvec_list, obj_list, num_samples_run, exit_info = self.evaluate_objective(x, number_of_samples, params)

        # Handle exit conditions (f < min obj value or maxfun reached)
        if exit_info is not None:
            if num_samples_run > 0:
                self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                      x_in_abs_coords=True)
            return exit_info  # didn't fix geometry - return & quit

        # Otherwise, add new results
        self.model.change_point(knew, xnew, rvec_list[0, :], self.nx)  # expect step, not absolute x
        for i in range(1, num_samples_run):
            self.model.add_new_sample(knew, rvec_extra=rvec_list[i, :])

        # Estimate actual reduction to add to diffs vector
        obj = sumsq(np.mean(rvec_list[:num_samples_run, :], axis=0)) # estimate actual objective value
        # pred_reduction = - calculate_model_value(gopt, H, d)
        pred_reduction = - model_value(gopt, H, d)
        if self.h is not None:
            obj += self.h(remove_scaling(x, self.scaling_changes), *self.argsh)
            # since m(0) = h(x)
            pred_reduction = self.h(remove_scaling(x, self.scaling_changes), *self.argsh) - model_value(gopt, H, d, x, self.h, self.argsh, self.scaling_changes)
        actual_reduction = objopt - obj
        self.diffs = [abs(pred_reduction - actual_reduction), self.diffs[0], self.diffs[1]]
        return None  # exit_info = None

    def check_and_fix_geometry(self, distsq_thresh, update_delta, number_of_samples, params):  # check and fix geometry
        # Find the point furthest from xopt
        sq_distances = self.model.distances_to_xopt()
        knew = np.argmax(sq_distances)
        distsq = sq_distances[knew]

        if distsq <= distsq_thresh:  # points close enough - don't do anything
            return False, None  # didn't fix geometry at all

        dist = sqrt(distsq)
        if update_delta:  # optional
            self.delta = max(min(0.1 * self.delta, 0.5 * dist), 1.5 * self.rho)

        adelt = max(min(0.1 * dist, self.delta), self.rho)
        dsq = adelt ** 2
        if dsq <= params("general.rounding_error_constant") * sumsq(self.model.xopt()):
            self.model.shift_base(self.model.xopt())

        exit_info = self.geometry_step(knew, adelt, number_of_samples, params)
        return True, exit_info  # didn't fix geometry - return & continue

    def evaluate_objective(self, x, number_of_samples, params):
        # Sample from objective function several times, keeping track of maxfun and min_obj_value throughout
        rvec_list = np.zeros((number_of_samples, self.m()))
        obj_list = np.zeros((number_of_samples,))
        num_samples_run = 0
        incremented_nx = False
        exit_info = None

        for i in range(number_of_samples):
            if self.nf >= self.maxfun:
                exit_info = ExitInformation(EXIT_MAXFUN_WARNING, "Objective has been called MAXFUN times")
                break  # quit

            self.nf += 1
            if not incremented_nx:
                self.nx += 1
                incremented_nx = True
            rvec_list[i, :], obj_list[i] = eval_least_squares_with_regularisation(self.objfun, remove_scaling(x, self.scaling_changes), self.h,
                                            argsf=self.argsf, argsh=self.argsh, verbose=self.do_logging, eval_num=self.nf, pt_num=self.nx,
                                            full_x_thresh=params("logging.n_to_print_whole_x_vector"),
                                            check_for_overflow=params("general.check_objfun_for_overflow"))
            num_samples_run += 1

        # Check if the average value was below our threshold
        # QUESTION: how to choose x in h when using averaged values
        if self.h is None:
            if num_samples_run > 0 and \
                            sumsq(np.mean(rvec_list[:num_samples_run, :], axis=0)) <= self.model.min_objective_value():
                exit_info = ExitInformation(EXIT_SUCCESS, "Objective is sufficiently small")
        else:
            if num_samples_run > 0 and \
                            sumsq(np.mean(rvec_list[:num_samples_run, :], axis=0)) + self.h(remove_scaling(x, self.scaling_changes),*self.argsh) <= self.model.min_objective_value():
                exit_info = ExitInformation(EXIT_SUCCESS, "Objective is sufficiently small")

        return rvec_list, obj_list, num_samples_run, exit_info

    def choose_point_to_replace(self, d, skip_kopt=True):
        delsq = self.delta ** 2
        scaden = None
        knew = None  # may knew never be set here?
        exit_info = None

        try:
            cs, gs = self.model.lagrange_gradient(k=None)  # find all Lagrange polynomials for k in range(self.model.npt())
        except LA.LinAlgError:
            exit_info = ExitInformation(EXIT_LINALG_ERROR, "Singular matrix when choosing point to replace")
            return knew, exit_info

        for k in range(self.model.npt()):
            if skip_kopt and k == self.model.kopt:
                continue  # skip this k

            # Extract Lagrange polynomial (based at xopt)
            c = cs[k]
            g = gs[:, k]

            den = c + np.dot(g, d)

            distsq = sumsq(self.model.xpt(k) - self.model.xopt())
            temp = max(1.0, (distsq / delsq) ** 2)
            if scaden is None or temp * abs(den) > scaden:
                scaden = temp * abs(den)
                knew = k

        return knew, exit_info

    def done_with_current_rho(self, xnew, gnew, crvmin, H, current_iter):
        # (xnew, gnew, crvmin) come from trust region step
        # H is Hessian of model for the full objective

        # Wait at least 3 iterations between reductions of rho
        if current_iter <= self.last_successful_iter + 2:
            return False

        errbig = max(self.diffs)
        frhosq = 0.125 * self.rho ** 2
        if crvmin > 0.0 and errbig > frhosq * crvmin:
            return False

        bdtol = errbig / self.rho
        for j in range(self.n()):
            bdtest = bdtol
            if xnew[j] == self.model.sl[j]:
                bdtest = gnew[j]
            if xnew[j] == self.model.su[j]:
                bdtest = -gnew[j]
            if bdtest < bdtol:
                curv = H[j,j]
                bdtest += 0.5 * curv * self.rho
                if bdtest < bdtol:
                    return False
        # Otherwise...
        return True

    def reduce_rho(self, current_iter, params):
        alpha1 = params("tr_radius.alpha1")
        alpha2 = params("tr_radius.alpha2")
        ratio = self.rho / self.rhoend
        if ratio <= 16.0:
            new_rho = self.rhoend
        elif ratio <= 250.0:
            new_rho = sqrt(ratio) * self.rhoend  # geometric average of rho and rhoend
        else:
            new_rho = alpha1 * self.rho

        self.delta = max(alpha2 * self.rho, new_rho)  # self.rho = old rho
        self.rho = new_rho
        self.last_successful_iter = current_iter  # reset successful iteration check
        return

    def calculate_ratio(self, x, current_iter, rvec_list, d, gopt, H):
        exit_info = None
        # estimate actual objective value
        obj = sumsq(np.mean(rvec_list, axis=0))
        # pred_reduction = - calculate_model_value(gopt, H, d)
        pred_reduction = - model_value(gopt, H, d)
        if self.h is not None:
            # QUESTION: x+d here correct? rvec_list takes mean value
            obj += self.h(remove_scaling(x+d, self.scaling_changes), *self.argsh) 
            # since m(0) = h(x)
            pred_reduction = self.h(remove_scaling(x, self.scaling_changes), *self.argsh) - model_value(gopt, H, d, x, self.h, self.argsh, self.scaling_changes)
        actual_reduction = self.model.objopt() - obj
        self.diffs = [abs(actual_reduction - pred_reduction), self.diffs[0], self.diffs[1]]
        if min(sqrt(sumsq(d)), self.delta) > self.rho:  # if ||d|| >= rho, successful!
            self.last_successful_iter = current_iter
        if pred_reduction < 0.0:
            if len(self.model.projections) > 1: # if we are using multiple projections, only warn since likely due to constraint intersection
                exit_info = ExitInformation(EXIT_TR_INCREASE_WARNING, "Either multiple constraints are active or trust region step gave model increase")
            else:
                exit_info = ExitInformation(EXIT_TR_INCREASE_ERROR, "Trust region step gave model increase")
        ratio = actual_reduction / pred_reduction
        return ratio, exit_info

    def terminate_from_slow_iterations(self, current_iter, params):
        if len(self.last_iters_step_taken) <= params("slow.history_for_slow"):
            # Not enough info, simply append
            self.last_iters_step_taken.append(current_iter)
            self.last_fopts_step_taken.append(self.model.objopt())
            this_iter_slow = False
        else:
            # Enough info - shift values
            self.last_iters_step_taken = self.last_iters_step_taken[1:] + [current_iter]
            self.last_fopts_step_taken = self.last_fopts_step_taken[1:] + [self.model.objopt()]
            this_iter_slow = (log(self.last_fopts_step_taken[0]) - log(self.model.objopt())) / \
                             float(params("slow.history_for_slow")) < params("slow.thresh_for_slow")
        # Update counter of number of slow iterations
        if this_iter_slow:
            self.num_slow_iters += 1
            if self.do_logging:
                module_logger.info("Slow iteration (%g consecutive so far, max allowed %g)"
                             % (self.num_slow_iters, params("slow.max_slow_iters")))
        else:
            self.num_slow_iters = 0
            if self.do_logging:
                module_logger.debug("Non-slow iteration")
        return this_iter_slow, self.num_slow_iters >= params("slow.max_slow_iters")

    def soft_restart(self, number_of_samples, nruns_so_far, params, x_in_abs_coords_to_save=None, rvec_to_save=None,
                     nsamples_to_save=None):
        # A successful run is one where we reduced fopt
        if self.model.objopt() < self.last_run_fopt:
            self.last_successful_run = nruns_so_far
        self.last_run_fopt = self.model.objopt()

        ok_to_do_restart = (nruns_so_far - self.last_successful_run < params("restarts.max_unsuccessful_restarts")) and \
                           (self.nf < self.maxfun)

        if not ok_to_do_restart:
            # last outputs are (exit_flag, exit_str, return_to_new_tr_iteration)
            exit_info = ExitInformation(EXIT_MAXFUN_WARNING, "Objective has been called MAXFUN times")
            if nruns_so_far - self.last_successful_run >= params("restarts.max_unsuccessful_restarts"):
                exit_info = ExitInformation(EXIT_SUCCESS, "Reached maximum number of unsuccessful restarts")
            return exit_info

        # Save current best point and the one the user wanted to save too
        if x_in_abs_coords_to_save is not None:
            assert rvec_to_save is not None, "Soft restart: specified x_to_save but not rvec_to_save"
            assert nsamples_to_save is not None, "Soft restart: specified x_to_save but not nsamples_to_save"
            self.model.save_point(x_in_abs_coords_to_save, rvec_to_save, nsamples_to_save, self.nx, x_in_abs_coords=True)
        self.model.save_point(self.model.xopt(abs_coordinates=True), self.model.ropt(), self.nx,
                              self.model.nsamples[self.model.kopt], x_in_abs_coords=True)

        if self.do_logging:
            module_logger.info("Soft restart [currently, f = %g after %g function evals]" % (self.model.objopt(), self.nf))
        # Resetting method: reset delta and rho, then move the closest 'num_steps' points to xk to improve geometry
        # Note: closest points because we are suddenly increasing delta & rho, so we want to encourage spreading out points
        self.delta = self.rhobeg
        self.rho = self.rhobeg
        self.diffs = [0.0, 0.0, 0.0]
        
        # Forget history of slow iterations
        self.last_iters_step_taken = []
        self.last_fopts_step_taken = []
        self.num_slow_iters = 0

        all_sq_dist = self.model.distances_to_xopt()[:self.model.npt()]
        closest_points = np.argsort(all_sq_dist)
        if params("restarts.soft.move_xk"):
            closest_points = closest_points[:params("restarts.soft.num_geom_steps")]
            upper_limit = self.model.num_pts
        else:
            closest_points = closest_points[1:params("restarts.soft.num_geom_steps")+1]
            upper_limit = self.model.num_pts - 1

        for i in range(min(params("restarts.soft.num_geom_steps"), upper_limit)):
            # Determine which point to update (knew)
            knew = closest_points[i]

            # Using adelt=delta in fix_geometry (adelt determines the ball to max lagrange poly in altmov)
            # [Only reason actual 'delta' is needed in fix_geometry is for calling nsamples()]
            exit_info = self.geometry_step(knew, self.delta, number_of_samples, params)
            if exit_info is not None:
                return exit_info

        # Now increase npt, if required
        if params("restarts.increase_npt") and self.model.npt() < params("restarts.max_npt"):
            num_pts_to_add = min(params("restarts.increase_npt_amt"), params("restarts.max_npt") - self.model.npt())
            # First n points will be random orthogonal directions; the rest will be purely random directions
            # sl <= xopt + dirn <= su   -or equivalently-   sl-xopt <= dirn <= su-xopt
            xopt = self.model.xopt()
            dirns = random_directions_within_bounds(num_pts_to_add, self.delta, self.model.sl - xopt,
                                                    self.model.su - xopt)
            for i in range(num_pts_to_add):
                xnew = self.model.xopt() + dirns[i, :]  # always base move around best value so far
                x = self.model.as_absolute_coordinates(xnew)
                rvec_list, obj_list, num_samples_run, exit_info = self.evaluate_objective(x, number_of_samples, params)

                # Handle exit conditions (f < min obj value or maxfun reached)
                if exit_info is not None:
                    if num_samples_run > 0:
                        self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                              x_in_abs_coords=True)
                    return exit_info  # return & quit

                # Otherwise, add new results
                self.model.add_new_point(xnew, rvec_list[0, :], self.nx)  # expect step, not absolute x
                for i in range(1, num_samples_run):
                    self.model.add_new_sample(self.model.npt() - 1, rvec_extra=rvec_list[i, :])

            if self.do_logging:
                module_logger.info("Soft restart: added %g new directions, npt is now %g" % (num_pts_to_add, self.model.npt()))

        # Otherwise, we are doing a restart
        self.last_successful_iter = 0
        return None  # exit_info = None

    def move_furthest_points(self, number_of_samples, num_pts_to_move, params):
        # get points furthest from (new) xopt and move to a good geometry point
        all_sq_dist = self.model.distances_to_xopt()[:self.model.npt()]
        furthest_points = np.argsort(all_sq_dist)[::-1]  # indices from furthest to closest (last is kopt)

        for i in range(min(num_pts_to_move, len(furthest_points) - 1)):
            # Determine which point to update (knew)
            knew = furthest_points[i]

            # Using adelt=delta in fix_geometry (adelt determines the ball to max lagrange poly in altmov)
            # [Only reason actual 'delta' is needed in fix_geometry is for calling nsamples()]
            adelt = self.delta
            exit_info = self.geometry_step(knew, adelt, number_of_samples, params)

            if exit_info is not None:
                return exit_info

        return None

    def all_values_within_noise_level(self, params):
        all_fvals_within_noise = True
        # We scale the expected noise by sqrt(# samples used to generate averages)
        if params("noise.additive_noise_level") is not None:
            add_noise = params("noise.scale_factor_for_quit") * params("noise.additive_noise_level")
            for k in range(self.model.npt()):
                all_fvals_within_noise = all_fvals_within_noise and \
                                (self.model.objval[k] <= self.model.objopt() + add_noise / sqrt(self.model.nsamples[k]))
        else:  # noise_level_multiplicative
            ratio = 1.0 + params("noise.scale_factor_for_quit") * params("noise.multiplicative_noise_level")
            for k in range(self.model.npt()):
                this_ratio = self.model.objval[k] / self.model.objopt()  # fval_opt strictly positive (would have quit o/w)
                all_fvals_within_noise = all_fvals_within_noise and (
                    this_ratio <= ratio / sqrt(self.model.nsamples[k]))
        return all_fvals_within_noise

    def move_furthest_points_momentum(self, d, number_of_samples, num_pts_to_move, params):
        all_sq_dist = self.model.distances_to_xopt()[:self.model.npt()]
        furthest_points = np.argsort(all_sq_dist)[::-1]  # indices from furthest to closest (last is kopt)

        xopt = self.model.xopt()
        dirns = random_directions_within_bounds(num_pts_to_move, self.delta, self.model.sl - xopt, self.model.su - xopt)

        for i in range(min(num_pts_to_move, len(furthest_points) - 1)):
            # Determine which point to update (knew)
            knew = furthest_points[i]
            if np.dot(dirns[i, :], d) < 0.0:
                dirns[i, :] = -dirns[i, :]  # flip so pointing in direction
                flipped = True
            else:
                flipped = False
            # However, this may send us outside our bounds if we are near the boundary; don't flip if the flip
            # reduces the size of the step by too much
            if flipped:
                fixed_dirn = np.maximum(np.minimum(dirns[i, :], self.model.su - self.model.xopt()),
                                         self.model.sl - self.model.xopt())
                if np.linalg.norm(fixed_dirn) < 1e-3 * self.delta:
                    dirns[i, :] = -dirns[i, :]
            xnew = np.maximum(np.minimum(self.model.xopt() + dirns[i, :], self.model.su), self.model.sl)
            x = self.model.as_absolute_coordinates(xnew)
            rvec_list, obj_list, num_samples_run, exit_info = self.evaluate_objective(x, number_of_samples, params)

            # Handle exit conditions (f < min obj value or maxfun reached)
            if exit_info is not None:
                if num_samples_run > 0:
                    self.model.save_point(x, np.mean(rvec_list[:num_samples_run, :], axis=0), num_samples_run, self.nx,
                                          x_in_abs_coords=True)
                return exit_info  # return & quit

            # Otherwise, add new results
            self.model.change_point(knew, xnew, rvec_list[0, :], self.nx)  # expect step, not absolute x
            for i in range(1, num_samples_run):
                self.model.add_new_sample(knew, rvec_extra=rvec_list[i, :])
        return None

